import theano.tensor as T
import numpy as np

__all__ = ['var']


def var(name, label=None, observed=False, const=False, vector=False, lower=None, upper=None):
    if vector and not observed:
        raise ValueError('Currently, only observed variables can be vectors')

    if observed and const:
        raise ValueError('Observed variables are automatically const')

    if vector:
        var = T.vector(name)
    else:
        var = T.scalar(name)

    var._name = name
    var._label = label
    var._observed = observed
    var._const = observed or const
    var._lower = lower or -np.inf
    var._upper = upper or np.inf

    return var



import numpy as np
import theano.tensor as T

from mle.model import Model

__all__ = ['Uniform', 'Normal', 'Mix2', 'Join']


def alltrue(conditionals):
    ret = 1
    for condition in conditionals:
        ret *= condition
    return ret


def bound(logp, *conditionals):
    return T.switch(alltrue(conditionals), logp, -np.inf)


class Uniform(Model):
    def __init__(self, x, lower, upper, *args, **kwargs):
        super(Uniform, self).__init__(*args, **kwargs)
        self._logp = T.log(T.switch(
            T.gt(x, upper), 0,
            T.switch(T.lt(x, lower), 0, 1/(upper-lower))
        ))
        self._cdf = T.switch(
            T.gt(x, upper), 1,
            T.switch(T.lt(x, lower), 0, (x-lower)/(upper-lower))
        )
        self._add_expr('x', x)
        self._add_expr('lower', lower)
        self._add_expr('upper', upper)


class Normal(Model):
    def __init__(self, x, mu, sigma, *args, **kwargs):
        super(Normal, self).__init__(*args, **kwargs)
        self._logp = bound(
            -(x - mu)**2 / (2 * sigma**2) + T.log(1 / T.sqrt(sigma**2 * 2*np.pi)),
            sigma > 0
        )
        self._cdf = 0.5 * (1 + T.erf((x - mu)/(sigma*T.sqrt(2))))
        self._add_expr('x', x)
        self._add_expr('mu', mu)
        self._add_expr('sigma', sigma)


class Mix2(Model):
    def __init__(self, theta, dist1, dist2, *args, **kwargs):
        super(Mix2, self).__init__(*args, **kwargs)
        self._logp = bound(
            T.log(theta*T.exp(dist1._logp) + (1-theta)*T.exp(dist2._logp)),
            theta >= 0, theta <= 1
        )
        self._cdf = lambda: self.theta*self.dist1._cdf + (1-self.theta)*self.dist2._cdf
        self._add_expr('theta', theta)
        self._add_submodel('dist1', dist1)
        self._add_submodel('dist2', dist2)


class Join(Model):
    def __init__(self, dist1, dist2=None, *args, **kwargs):
        super(Join, self).__init__(*args, **kwargs)
        self._add_submodel('dist1', dist1)
        if dist2 is None:
            self._logp = T.sum(dist1._logp)
        else:
            self._add_submodel('dist2', dist2)
            self._logp = dist1._logp + dist2._logp



import logging
from time import clock

import numpy as np
from scipy.optimize import minimize
from theano import function, gof, shared, config
import theano.tensor as T

from mle.util import memoize

__all__ = ['Model']


class Model(object):
    def __init__(self):
        self.exprs = dict()
        self.submodels = dict()

    def fit(self, data, init, method='BFGS', verbose=False):
        data_args = []
        shared_params = []
        for var in self.observed:
            try:
                data[var.name]
            except KeyError:
                raise ValueError('Random variable {} required by model not found in dataset'
                                 .format(var.name))
            except:
                raise ValueError('The fitted dataset must support string indexing')

            data_args.append(np.array(data[var.name]))
            shared_params.append((var, shared(data[var.name].astype(config.floatX), borrow=True)))

        const = []
        x0 = []
        for par in self.parameters:
            if par.name not in init:
                raise ValueError('No initial value specified for Parameter {}'.format(par.name))
            if par._const:
                const.append(init[par.name])
            else:
                x0.append(init[par.name])

        logp = function(self.constant + self.floating, -T.sum(self._logp),
                        givens=shared_params, allow_input_downcast=True)
        g_logp = function(self.constant + self.floating, T.grad(-T.sum(self._logp), self.floating),
                          givens=shared_params, allow_input_downcast=True)

        def func(pars):
            return logp(*(const + list(pars)))

        def g_func(pars):
            return np.array(g_logp(*(const + list(pars))))

        logging.info('Minimizing negative log-likelihood of model...')

        start = clock()
        if method.upper() == 'MINUIT':
            from .minuit import fmin_minuit
            results = fmin_minuit(func, x0, map(str, self.floating), verbose=verbose)
        else:
            results = minimize(func, method=method, jac=g_func, x0=x0, options={'disp': True})
            names = [x.name for x in self.parameters]
            results.x = {n: x for n, x in zip(names, results.x)}
        fit_time = clock() - start

        # Add constant parameters to results
        for par in self.parameters:
            if par._const:
                results.x[par.name] = init[par.name]

        results['fit_time'] = fit_time

        return results

    def _add_expr(self, name, expr):
        self.exprs[name] = expr

    def _add_submodel(self, name, model):
        self.submodels[name] = model

    def _add_compiled_expr(self, name, expr):

        @memoize
        def compiler():
            logging.info('Compiling {}...'.format(name))
            return function(self.observed + self.parameters, expr, allow_input_downcast=True)

        setattr(self, 'compile_' + name, compiler)

        def compiled(model, *args):
            compiler = getattr(self, 'compile_' + name)
            func = compiler()
            return func(*args)

        setattr(self, name, compiled)

    @property
    def variables(self):
        return filter(lambda x: isinstance(x, T.TensorVariable), gof.graph.inputs([self._logp]))

    @property
    def observed(self):
        return list(filter(lambda x: x._observed, self.variables))

    @property
    def parameters(self):
        return list(filter(lambda x: not x._observed, self.variables))

    @property
    def constant(self):
        return list(filter(lambda x: not x._observed and x._const, self.variables))

    @property
    def floating(self):
        return list(filter(lambda x: not x._const, self.variables))



try:
    from iminuit import Minuit
except ImportError:
    raise ImportError("The iminuit package must be installed in order to use `method='MINUIT'`")

from iminuit.util import make_func_code
from scipy.optimize import OptimizeResult


class Min_Func:

    def __init__(self, f, names):
        self.f = f
        self.func_code = make_func_code(names)
        self.func_defaults = None

    def __call__(self, *args):
        return self.f(args)


def fmin_minuit(func, x0, names=None, verbose=False):
    inits = dict()

    if verbose:
        print_level = 2
    else:
        print_level = 0

    if names is None:
        names = map(lambda x: 'param' + str(x), range(len(x0)))
    else:
        assert(len(x0) == len(names))

    for n, x in zip(names, x0):
        inits[n] = x
        # TODO use a method to set this correctly
        inits['error_' + n] = 1

    m = Minuit(Min_Func(func, names), print_level=print_level, errordef=1, **inits)
    a, b = m.migrad()

    return OptimizeResult(
        x=m.values,
        fun=a['fval'],
        edm=a['edm'],
        nfev=a['nfcn'],
        is_valid=a['is_valid'],
        has_valid_parameters=a['has_valid_parameters'],
    )



import functools

from theano import Variable, scan
from theano.tensor import arange, grad, stack
from theano.gradient import format_as


def hessian_(cost, wrt, consider_constant=None, disconnected_inputs='raise'):
    """

    :type cost: Scalar (0-dimensional) Variable.
    :type wrt: Vector (1-dimensional tensor) 'Variable' or list of
               vectors (1-dimensional tensors) Variables
    :param consider_constant: a list of expressions not to backpropagate through
    :type disconnected_inputs: string
    :param disconnected_inputs: Defines the behaviour if some of the variables
        in ``wrt`` are not part of the computational graph computing ``cost``
        (or if all links are non-differentiable). The possible values are:
        - 'ignore': considers that the gradient on these parameters is zero.
        - 'warn': consider the gradient zero, and print a warning.
        - 'raise': raise an exception.
    :return: either a instance of Variable or list/tuple of Variables
            (depending upon `wrt`) repressenting the Hessian of the `cost`
            with respect to (elements of) `wrt`. If an element of `wrt` is not
            differentiable with respect to the output, then a zero
            variable is returned. The return value is of same type
            as `wrt`: a list/tuple or TensorVariable in all cases.
    """
    # Check inputs have the right format
    assert isinstance(cost, Variable), "tensor.hessian expects a Variable as `cost`"
    assert cost.ndim == 0, "tensor.hessian expects a 0 dimensional variable as `cost`"

    using_list = isinstance(wrt, list)
    using_tuple = isinstance(wrt, tuple)

    if isinstance(wrt, (list, tuple)):
        wrt = list(wrt)
    else:
        wrt = [wrt]

    expr = grad(cost, wrt, consider_constant=consider_constant,
                disconnected_inputs=disconnected_inputs)

    # It is possible that the inputs are disconnected from expr,
    # even if they are connected to cost.
    # This should not be an error.
    hess, updates = scan(
        lambda i, y: grad(y[i], wrt, consider_constant=consider_constant,
                          disconnected_inputs='ignore'),
        sequences=arange(len(expr)), non_sequences=[expr]
    )
    assert not updates, ("Scan has returned a list of updates. This should not happen! Report "
                         "this to theano-users (also include the script that generated the error)")

    return format_as(using_list, using_tuple, stack(hess)[0])


def memoize(obj):
    """An expensive memoizer that works with unhashables."""
    cache = obj.cache = {}

    @functools.wraps(obj)
    def memoizer(*args, **kwargs):
        key = (hashable(args), hashable(kwargs))

        if key not in cache:
            cache[key] = obj(*args, **kwargs)

        return cache[key]
    return memoizer


def hashable(a):
    """Turn some unhashable objects into hashable ones."""
    if isinstance(a, dict):
        return hashable(a.items())
    try:
        return tuple(map(hashable, a))
    except:
        return a



from setuptools import setup

setup(name='sengbao',
      version='0.1',
      description='High Performance Maximum Likelihood Estimations',
      url='http://github.com/ibab/python-mle',
    #   author='Igor Babuschkin',
    #   author_email='igor@babuschk.in',
      license='MIT',
      packages=['mle', 'mle.distributions'],
      install_requires=[
          'numpy',
          'scipy',
          'theano',
      ],
      zip_safe=False)
